@@ -1,17 +1,20 @@
 # create logger
 import json
-from typing import Optional, List
+from pathlib import Path
+from typing import Optional, List, Tuple
 
 import typer
 from bigeye_sdk.client.enum import Method
 
 from bigeye_cli.enums import OPS_METRICS
-from bigeye_sdk.model.protobuf_enum_facade import MetricStatus, SimpleMetricCategory
+from bigeye_sdk.model.protobuf_enum_facade import MetricStatus, SimpleMetricCategory, SimpleMetricTemplateParameterType, \
+    SimpleMetricTemplateFieldType
 from bigeye_sdk.functions.aws import get_secret
 from bigeye_sdk.generated.com.bigeye.models.generated import MetricInfoList, WarehouseType, CreateSourceRequest, \
-    TimeIntervalType, PredefinedMetricName, MetricCreationState, MetricConfiguration, ForecastModelType
+    TimeIntervalType, PredefinedMetricName, MetricCreationState, MetricConfiguration, ForecastModelType, Source, Table
 from bigeye_sdk.log import get_logger
-from bigeye_cli.functions import write_metric_info, cli_client_factory, write_table_info, write_debug_queries
+from bigeye_cli.functions import write_metric_info, cli_client_factory, write_table_info, write_debug_queries, \
+    write_metric_templates
 from bigeye_sdk.functions.metric_functions import has_auto_threshold
 from bigeye_sdk.functions.core_py_functs import int_enum_enum_list_joined
 
@@ -31,7 +34,7 @@ def regen_autometrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -71,7 +74,7 @@ def deploy_all_autometrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -134,7 +137,7 @@ def delete_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -183,7 +186,7 @@ def rebuild(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             ...
             , "--warehouse_id"
@@ -208,7 +211,7 @@ def get_table_info(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -253,7 +256,7 @@ def get_metric_info(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -310,7 +313,7 @@ def run_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -331,7 +334,7 @@ def run_metrics(
             None
             , "--table_name"
             , "-tn"
-            , help="Table Namess. E.g. -tn some_table -tn some_other_table")
+            , help="Table Names. E.g. -tn some_table -tn some_other_table")
 ):
     """Runs all metrics for a warehouse, particular schemas in a warehouse, or tables by id."""
     client = cli_client_factory(bigeye_conf)
@@ -360,7 +363,7 @@ def add_source(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         secret_name: str = typer.Option(
             None
             , "--secret_name"
@@ -402,7 +405,7 @@ def delete_source(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -421,7 +424,7 @@ def backfill_autothresholds(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -458,7 +461,7 @@ def backfill_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -508,7 +511,7 @@ def run_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -547,7 +550,7 @@ def schedule_all_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -593,7 +596,7 @@ def unschedule_all_metrics(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -640,7 +643,7 @@ def set_metric_time(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -660,7 +663,7 @@ def set_metric_time(
             ...
             , "--metric_column_name"
             , "-cn"
-            , help="Possible metric column names.."),
+            , help="Possible metric column names."),
         replace_set_metric_times: bool = typer.Option(
             False
             , "--replace"
@@ -673,15 +676,15 @@ def set_metric_time(
     client = cli_client_factory(bigeye_conf)
 
     if warehouse_id and table_ids:
-        raise Exception('Must include either table IDs OR warehouse id OR schema id.')
-    elif warehouse_id:
+        raise Exception('Must include either table IDs OR warehouse ID OR schema ID.')
+    elif warehouse_id or schema_id:
         client.set_source_metric_times(column_names=metric_column_names, wid=warehouse_id, sid=schema_id,
                                        replace=replace_set_metric_times)
     elif table_ids:
         client.set_table_metric_times(column_names=metric_column_names, table_ids=table_ids,
                                       replace=replace_set_metric_times)
     else:
-        raise Exception('Must include either table IDs OR warehouse id.')
+        raise Exception('Must include either table IDs OR warehouse ID OR schema ID.')
 
 
 @app.command()
@@ -690,7 +693,7 @@ def unset_metric_time(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -721,7 +724,7 @@ def get_metric_queries(
             None
             , "--bigeye_conf"
             , "-b"
-            , help="Bigeye Basic Auth Configuration File"),
+            , help="Bigeye Basic Auth Configuration"),
         warehouse_id: int = typer.Option(
             None
             , "--warehouse_id"
@@ -757,3 +760,188 @@ def get_metric_queries(
     r = client.get_debug_queries(metric_ids=mids)
 
     write_debug_queries(output_path=output_path, queries=r)
+
+
+@app.command()
+def upsert_template(
+        bigeye_conf: str = typer.Option(
+            None
+            , "--bigeye_conf"
+            , "-b"
+            , help="Bigeye Basic Auth Configuration"),
+        source_name: str = typer.Option(
+            ...
+            , "--source_name"
+            , "-sn"
+            , help="The name of the source where the template will be defined"),
+        template_name: str = typer.Option(
+            ...
+            , "--name"
+            , "-tn"
+            , help="The user defined name of the template"),
+        template_body: str = typer.Option(
+            ...
+            , "--body"
+            , "-tb"
+            , help="The string to define the template"),
+        return_type: SimpleMetricTemplateFieldType = typer.Option(
+            ...
+            , "--returns"
+            , "-rt"
+            , help="The data type returned by the template; i.e. NUMERIC, BOOLEAN"),
+        # TODO: Update to List[Tuple[str, SimpleMetricTemplateParameterType]] when supported by typer.
+        parameters: List[str] = typer.Option(
+            ...
+            , "--params"
+            , "-p"
+            , help="A list of key/value pairs for parameters; ex. -p column=COLUMN_REFERENCE -p table=STRING"),
+        template_id: Optional[int] = typer.Option(
+            0
+            , "--template_id"
+            , "-id"
+            , help="The template ID (Only required when updating a template)."
+        )
+):
+    """Create or update a template for a source."""
+    # TODO: Lookup template ID, once template names are unique
+    client = cli_client_factory(bigeye_conf)
+
+    source: Source = [s for s in client.get_sources().sources if s.name == source_name][0]
+    log.info(f"Creating metric template in warehouse {source.name}")
+    r = client.upsert_metric_template(id=template_id,
+                                      name=template_name,
+                                      template=template_body,
+                                      return_type=return_type,
+                                      parameters=parameters,
+                                      source=source)
+    log.info(f"Upsert for metric template complete:\n\tID:{r.metric_template.id}\n\tName:{r.metric_template.name}\n\t")
+
+
+@app.command()
+def get_all_metric_templates(
+        bigeye_conf: str = typer.Option(
+            None
+            , "--bigeye_conf"
+            , "-b"
+            , help="Bigeye Basic Auth Configuration"),
+        page_size: Optional[int] = typer.Option(
+            0
+            , "--page_size"
+            , "-ps"
+            , help="How many results to return per page"
+        ),
+        search: Optional[str] = typer.Option(
+            ""
+            , "--search"
+            , "-s"
+            , help="A search string to narrow results"
+        ),
+        output_path: str = typer.Option(
+            None
+            , "--output_path"
+            , "-op"
+            , help="The path to output templates"
+        ),
+        file_name: Optional[str] = typer.Option(
+            None
+            , "--file_name"
+            , "-fn"
+            , help="User defined file name"
+        )
+):
+    """Retrieve all metric templates and output to a file."""
+    client = cli_client_factory(bigeye_conf)
+    metric_templates = client.get_all_metric_templates(page_size=page_size, search=search)
+    if not output_path:
+        output_path = Path.cwd()
+
+    write_metric_templates(output_path=output_path, metric_templates=metric_templates, file_name=file_name)
+
+
+@app.command()
+def delete_template(
+        bigeye_conf: str = typer.Option(
+            None
+            , "--bigeye_conf"
+            , "-b"
+            , help="Bigeye Basic Auth Configuration"),
+        template_id: int = typer.Option(
+            ...
+            , "--template_id"
+            , "-tid"
+            , help="The ID of the metric template")
+):
+    """Delete a template."""
+    client = cli_client_factory(bigeye_conf)
+    client.delete_metric_template(template_id=template_id)
+    log.info(f"Template ID {template_id} deleted.")
+
+
+@app.command()
+def upsert_virtual_table(
+        bigeye_conf: str = typer.Option(
+            None
+            , "--bigeye_conf"
+            , "-b"
+            , help="Bigeye Basic Auth Configuration"),
+        virtual_table_name: str = typer.Option(
+            ...
+            , "--table_name"
+            , "-vtn"
+            , help="The name of the virtual table"),
+        sql: str = typer.Option(
+            ...
+            , "--sql"
+            , "-s"
+            , help="The SQL to define the table"),
+        source_name: str = typer.Option(
+            ...
+            , "--source_name"
+            , "-sn"
+            , help="The name of the source where the virtual table will be defined"),
+        update: bool = typer.Option(
+            False
+            , "--update"
+            , "-u"
+            , help="Create if false. Update if true.")
+):
+    """Create or update a virtual table."""
+    client = cli_client_factory(bigeye_conf)
+
+    source: Source = [s for s in client.get_sources().sources if s.name == source_name][0]
+
+    if update:
+        table: Table = [vt for vt in client.get_all_virtual_tables() if vt.table.warehouse_id == source.id and
+                        vt.table.name == virtual_table_name][0].table
+        log.info(f"Update {table.name} in warehouse {source.name}")
+        client.update_virtual_table(id=table.id, name=virtual_table_name, sql=sql, warehouse_id=source.id)
+    else:
+        log.info(f"Creating virtual table {virtual_table_name} in warehouse {source.name}")
+        client.create_virtual_table(name=virtual_table_name, sql=sql, warehouse_id=source.id)
+
+
+@app.command()
+def delete_virtual_table(
+        bigeye_conf: str = typer.Option(
+            None
+            , "--bigeye_conf"
+            , "-b"
+            , help="Bigeye Basic Auth Configuration"),
+        virtual_table_name: str = typer.Option(
+            ...
+            , "--table_name"
+            , "-vtn"
+            , help="The name of the virtual table"),
+        source_name: str = typer.Option(
+            ...
+            , "--source_name"
+            , "-sn"
+            , help="The name of the source where the template will be defined")
+):
+    """Delete a virtual table."""
+    client = cli_client_factory(bigeye_conf)
+    source: Source = [s for s in client.get_sources().sources if s.name == source_name][0]
+    table: Table = [vt for vt in client.get_all_virtual_tables() if vt.table.warehouse_id == source.id and
+                    vt.table.name == virtual_table_name][0].table
+
+    client.delete_virtual_table(table_id=table.id)
