@@ -7,17 +7,15 @@ import smart_open
 import typer
 
 from bigeye_cli.functions import cli_client_factory, write_debug_queries
-from bigeye_sdk.client.datawatch_client import BasicAuthDatawatchClient
-from bigeye_sdk.generated.com.bigeye.models.generated import TableList, Table, MetricConfiguration, MetricInfo
+from bigeye_sdk.generated.com.bigeye.models.generated import TableList, Table, MetricInfo
 from bigeye_sdk.model.sla_models import SlaMetrics
-from bigeye_cli import HOME_DIR
 
 from bigeye_sdk.log import get_logger
 from bigeye_sdk.functions.file_functs import serialize_listdict_to_json_file
 
 log = get_logger(__file__)
 
-app = typer.Typer(no_args_is_help=True, help='SLA Commands for Bigeye CLI')
+app = typer.Typer(no_args_is_help=True, help='SLA Commands for Bigeye CLI (use collections command instead)')
 
 
 @app.command()
@@ -41,7 +39,6 @@ def add_metric(
     """Add metric to an SLA."""
     log.info(f'Adding metric to SLA.')
     log.info(f'Bigeye API Configuration: {bigeye_conf} | Metric IDs: {metric_ids} | SLA ID: {sla_id}')
-
     client = cli_client_factory(bigeye_conf)
 
     client.upsert_metric_to_collection(add_metric_ids=metric_ids, collection_id=sla_id)
@@ -126,13 +123,13 @@ def migrate_from_json(bigeye_conf: str = typer.Option(
 
         # Clearing workspace specific settings.
         if not keep_notifications:
-            sla_metrics.sla.id = None
-        sla_metrics.sla.entity_info = None
-        sla_metrics.sla.owner = None
+            sla_metrics.collection.id = None
+        sla_metrics.collection.entity_info = None
+        sla_metrics.collection.owner = None
         if not keep_ids:
-            sla_metrics.sla.metric_ids = []
+            sla_metrics.collection.metric_ids = []
         if keep_notifications:
-            sla_metrics.sla.notification_channels = []
+            sla_metrics.collection.notification_channels = []
         m: MetricInfo
         for m in sla_metrics.metrics.metrics:
             if not keep_ids:
@@ -171,13 +168,13 @@ def migrate_from_json(bigeye_conf: str = typer.Option(
                     description=mc.description,
                 )
 
-                sla_metrics.sla.metric_ids.append(rmc.id)
+                sla_metrics.collection.metric_ids.append(rmc.id)
 
             except Exception as e:
-                log.exception(f"Error for SLA: {sla_metrics.sla.name}")
+                log.exception(f"Error for SLA: {sla_metrics.collection.name}")
                 log.exception(e)
 
-        client.create_collection_dep(sla_metrics.sla)
+        client.create_collection_dep(sla_metrics.collection)
 
 
 @app.command()
